/*************************************************************************\
**  source       * ssdir.c
**  directory    * ss
**  description  * Directory handling functions
**               * 
**               * 
**               * Copyright (C) 2006 Solid Information Technology Ltd
\*************************************************************************/
/*
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; only under version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
*/

#ifdef DOCUMENTATION
**************************************************************************

Implementation:
--------------

Portable directory routines for MS-DOS, OS/2 and Unix.

Limitations:
-----------

None.

Error handling:
--------------

Asserts.

Objects used:
------------

None.

Preconditions:
-------------

None.

Multithread considerations:
--------------------------

None.

Example:
-------


**************************************************************************
#endif /* DOCUMENTATION */


#include "ssenv.h"
#include "ssstdio.h"

#ifndef SS_NT

#include "sstime.h"
#include "ssstring.h"
#include "ssmem.h"
#include "ssdebug.h"
/* #include <fcntl.h> */
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#include "ssstdlib.h"
#if !defined(SS_UNIX) && !defined(SS_VMS)
#  include <sys/utime.h>
#elif defined(SS_UNIX) && !defined(SS_CRX) && !defined(SS_OSE)
#  include <utime.h>
#elif !defined(SS_OSE)
#  include <time.h>
#endif
/* #include <sys/locking.h> */
/* #include <sys/utime.h> */
/* #include <io.h> */
#include "sschcvt.h"
#include "ssc.h"
#include "ssdir.h"
#include "ssmem.h"
#include "ssdebug.h"

#if defined(SS_OS2)

#  define INCL_DOSFILEMGR
#  include <os2.h>

#  define DIRMASK         "\\*.*"
#  define DIRATTR         (FILE_NORMAL | FILE_DIRECTORY)

struct SsDirStruct {

    HDIR handle;
    FILEFINDBUF findbuf;
    SsDirEntT dirent;

};

#elif defined(SS_DOS)

#  include <dos.h>

#  define DIRMASK         "\\*.*"
#  define DIRATTR         (0x16)

#  define MAXNAMLEN   13

   typedef struct {
        unsigned char   stuff[30];
        char            namebuffer[MAXNAMLEN];
   } DIR;

struct SsDirStruct {

    DIR dir;
    SsDirEntT dirent;

};

#elif defined(SS_UNIX)

#  include <dirent.h>

#  define DIRMASK         ""
#  define DIRATTR         (0x16)

struct SsDirStruct {

    DIR* dir;
    struct dirent *xdirent;
    SsDirEntT dirent;

};

#endif




/*##**********************************************************************\
 * 
 *		SsDirOpen
 * 
 * Opens a directory for reading
 * 
 * Parameters : 
 * 
 *	pathname - in, use
 *		directory to be opened
 *
 * Return value - give :
 *
 *      pointer to directory structure
 * 
 * Limitations  :
 * 
 * Globals used :
 */
SsDirT *SsDirOpen(pathname)
char *pathname;
{
        SsDirT *dir = SsMemAlloc(sizeof(SsDirT));
        int pathlen = strlen(pathname);
        char *path = SsMemAlloc(pathlen + strlen(DIRMASK) + 2);
        char* dirmask = DIRMASK;

        strcpy(path, pathname);
#if defined(DOS_OR_OS2)
        if (pathlen && path[pathlen-1] == dirmask[0]) {
            path[pathlen-1] = '\0';
        }
#endif
        strcat(path, DIRMASK);

        {
#if defined(SS_OS2)

            USHORT ret, count = 1;

            dir->handle = HDIR_CREATE;

            ret = DosFindFirst((PSZ)path, &dir->handle, DIRATTR,
                &dir->findbuf, sizeof(FILEFINDBUF), &count, 0);

            if (ret) {
                SsMemFree(dir);
                SsMemFree(path);
                return(NULL);
            }

            dir->dirent.filename = SsMemAlloc(dir->findbuf.cchName+1);
            dir->dirent.used = FALSE;

            strncpy(dir->dirent.filename, dir->findbuf.achName,
                dir->findbuf.cchName);

            dir->dirent.filename[dir->findbuf.cchName] = '\0';
#elif defined(SS_DOS)
            union REGS regs;

            regs.h.ah = 0x1A;               /* set disk transfer address */
            regs.x.dx=&(dir->dir);
            intdos(&regs, &regs);

            regs.h.ah = 0x4E;               /* find first file */
            regs.x.dx = path;
            regs.x.cx = DIRATTR;               /* file attributes */
            intdos(&regs, &regs);

            if (regs.x.cflag) {
                SsMemFree(dir);
                SsMemFree(path);
                return(NULL);
            }

            dir->dirent.filename = SsMemAlloc(strlen(dir->dir.namebuffer)+1);
            dir->dirent.used = FALSE;

            strcpy(dir->dirent.filename, dir->dir.namebuffer);
#else
	    dir->dir = opendir(path);
            dir->dirent.used = TRUE;

	    if (!dir->dir) {
                SsMemFree(dir);
                SsMemFree(path);
                return(NULL);
            }
#endif
        }

        SsMemFree(path);

        return(dir);
}


/*##**********************************************************************\
 * 
 *		SsDirRead
 * 
 * Reads the next file from a directory opened by SsDirOpen
 * 
 * Parameters : 
 * 
 *	dir - in out, use
 *		directory to be read
 *
 * 
 * Return value - ref : 
 * 
 *      pointer to a directory entry
 * 
 * Limitations  :
 * 
 * Globals used :
 */
SsDirEntT *SsDirRead(dir)
SsDirT *dir;
{
        ss_dassert(dir != NULL);
        if (dir->dirent.used) {
#if defined(SS_OS2)
            USHORT ret, count = 1;

	    if (dir->dirent.filename) {
 	        SsMemFree(dir->dirent.filename);
	    }
	    dir->dirent.filename = NULL;

            ret = DosFindNext(dir->handle, &dir->findbuf, 
                sizeof(FILEFINDBUF), &count);

            if (ret) {
                return(NULL);
            }

            dir->dirent.filename = SsMemAlloc(dir->findbuf.cchName+1);
            dir->dirent.used = FALSE;

            strncpy(dir->dirent.filename, dir->findbuf.achName,
                dir->findbuf.cchName);

            dir->dirent.filename[dir->findbuf.cchName] = '\0';
#elif defined(SS_DOS)
            union REGS regs;

	    if (dir->dirent.filename) {
 	        SsMemFree(dir->dirent.filename);
	    }
	    dir->dirent.filename = NULL;

            regs.h.ah = 0x1A;               /* set disk transfer address */
            regs.x.dx = &(dir->dir);
            intdos(&regs, &regs);

            regs.h.ah = 0x4F;               /* find next file */
            intdos(&regs, &regs);

            if (regs.x.cflag) {
                return(NULL);
            }

            dir->dirent.filename = SsMemAlloc(strlen(dir->dir.namebuffer)+1);
            dir->dirent.used = FALSE;

            strcpy(dir->dirent.filename, dir->dir.namebuffer);
#else /* UNIX begins */

#if defined(SS_LINUX) || defined(SS_SOLARIS)
#  define d_namlen d_reclen
#endif
	    dir->dirent.filename = NULL;
	    dir->xdirent = readdir(dir->dir);

	    if (!dir->xdirent) {
	        return(NULL);
	    }	        
            /* Ari changed Sep 16, 1998
             * All systems do not contain d_namlen, therefore I assume
             * that d_name is NULL terminated.  let's see if this works
             */
	    dir->dirent.filename = dir->xdirent->d_name;
            dir->dirent.used = FALSE;
#endif
        }

        dir->dirent.used = TRUE;

        return(&(dir->dirent));
}

/*##**********************************************************************\
 * 
 *		SsDirClose
 * 
 * Closes a directory opened by SsDirOpen
 * 
 * Parameters : 
 * 
 *	dir - in, take
 *		directory to be closed
 *
 * Return value :
 * 
 * Limitations  :
 * 
 * Globals used :
 */
void SsDirClose(dir)
SsDirT *dir;
{
#ifdef OS2
        DosFindClose(dir->handle);
#else
#ifdef DOS
	/* how does one close a directory in dos? */
#else
	closedir(dir->dir);
#endif
#endif
	if (dir->dirent.filename) {
	    SsMemFree(dir->dirent.filename);
	}
        SsMemFree(dir);
}

#endif /* !SS_NT */

